import numpy as np


def exploit(x, model, params):
    """
    Purely exploiting acquisition function

    Doesn't take any parameters (apart from the model).
    """
    m, s, dmdx, dsdx = model.predictive_m_s_grad(x)
    f_acqu = m
    df_acqu = dmdx
    scipygradient = np.asmatrix(df_acqu).transpose()
    return f_acqu, scipygradient
